/*
 *	Hippogryf -- Common definitions
 *
 *	(c) 2018 Martin Mares <mj@ucw.cz>
 */

using namespace std;

#include <cassert>
#include <cstdlib>
#include <exception>
#include <iostream>
#include <memory>
#include <stdexcept>
#include <string>
#include <vector>

typedef unsigned int uint;

#define UNUSED __attribute__((unused))

// insn.cc

class insn;
class state;

class insn_def {
public:
	const char *syntax;
	const char *display;
	int parse(const string s, insn *ins);	// 0=unmatched, 1=partial name match, 2=partial arg match, 3=full, -1=error
	string show(const insn *ins);
	string show_pretty(const insn *ins);
	void (*exec)(const insn *ins, state *st);
};

#define MAX_ARGS 3

class insn {
public:
	insn();
	insn_def *def;
	int nargs;
	int arg[MAX_ARGS];
	void add_arg(int x) { assert(nargs < MAX_ARGS); arg[nargs++] = x; }
	bool parse(const string s);
	int partial_parse(const string s);
};

// program.cc

#define NUM_ROWS 20
#define NUM_COLS 10
#define MAX_ARRAY_LEN 1000000
#define MAX_STACK_DEPTH 1000
#define MAX_RUN_TIME 100000000

class program {
public:
	program();
	insn ins[NUM_ROWS][NUM_COLS];
	void load(istream *in);
	void load(const string filename);
	void save(ostream *out);
	void save(const string filename);
	pair<int, int> labels[27];	// Label 26 = start; -1,-1 = undefined
	void resolve_labels();
};

class array_state {
public:
	vector<int> x;
};

class pc_state {
public:
	int i, j;
	int dir_i, dir_j;
	void step();
	void unstep();
	void face(int dir);		// 0=north, 1=east, 2=west, 3=south
	void turn(int dir);		// 0=left, 1=right, 2=back
};

class state {
public:
	const program *pgm;

	// Program counter and direction vector
	pc_state pc;

	// Registers
	int num_reg[26];
	shared_ptr<array_state> array_reg[26];
	void reg_set_num(int i, int x);
	void reg_set_array(int i, shared_ptr<array_state> a);
	void reg_new_array(int i, int len);
	int reg_get_num(int i);
	shared_ptr<array_state> reg_get_array(int i);
	bool reg_is_num(int i);
	bool reg_is_array(int i);
	void show_regs();

	// Return stack
	vector<pc_state> return_stack;

	// Running
	state(const program *_pgm);
	int run_time;
	int debug;
	void run_step();
	void run();

	// Input and output
	void read_input(string in);
	void write_output(string out);
	vector<pair<int,int>> output_spec;
};

class hippo_error : public runtime_error {
public:
	int i, j;
	hippo_error(string s) : runtime_error(s), i(-1), j(-1) { }
	hippo_error(int _i, int _j, string s) : runtime_error(s), i(_i), j(_j) { }
};

class hippo_halt : public exception {
};
