/*
 *	Hippogryf -- Instructions
 *
 *	(c) 2018 Martin Mares <mj@ucw.cz>
 */

#include "hippogryf.h"

#include <cctype>

// Implementation of instructions

#define EXEC(name) void exec_##name(const insn *ins UNUSED, state *st UNUSED)
#define ARG(i) ins->arg[i]
#define RR(i) get_arg(ins, st, i)
#define RRA(i) st->reg_get_array(ARG(i))
#define WR(i,x) st->reg_set_num(ARG(i), x)

static int get_arg(const insn *ins, state *st, int i)
{
	int a = ins->arg[i];
	if (a < 1000)
		return st->reg_get_num(a);
	else
		return a - 1000;
}

EXEC(nop)
{
	// Does nothing
}

EXEC(let)
{
	int a = ARG(1);
	if (a < 1000 && st->reg_is_array(a))
		st->reg_set_array(ARG(0), st->reg_get_array(a));
	else
	WR(0, RR(1));
}

EXEC(add)
{
	WR(0, RR(1) + RR(2));
}

EXEC(sub)
{
	WR(0, RR(1) - RR(2));
}

EXEC(mul)
{
	WR(0, RR(1) * RR(2));
}

EXEC(div)
{
	int y = RR(2);
	if (!y)
		throw hippo_error("Dělení nulou");
	WR(0, RR(1) / y);
}

EXEC(mod)
{
	int y = RR(2);
	if (!y)
		throw hippo_error("Dělení nulou");
	int mod = RR(1) % y;
	if (mod < 0)
		mod += y;
	WR(0, mod);
}

EXEC(append)
{
	WR(0, 10*RR(1) + RR(2));
}

EXEC(shl)
{
	int by = RR(2);
	if (by < 0 || by > 10)
		throw hippo_error("Posun mimo rozsah");
	int x = RR(1);
	for (int i=0; i<by; i++)
		x *= 10;
	WR(0, x);
}

EXEC(shr)
{
	int by = RR(2);
	if (by < 0 || by > 10)
		throw hippo_error("Posun mimo rozsah");
	int x = RR(1);
	for (int i=0; i<by; i++)
		x /= 10;
	WR(0, x);
}

EXEC(ifeq)
{
	if (!(RR(0) == RR(1)))
		st->pc.step();
}

EXEC(ifne)
{
	if (!(RR(0) != RR(1)))
		st->pc.step();
}

EXEC(iflt)
{
	if (!(RR(0) < RR(1)))
		st->pc.step();
}

EXEC(ifgt)
{
	if (!(RR(0) > RR(1)))
		st->pc.step();
}

EXEC(turn)
{
	st->pc.turn(ARG(0));
}

EXEC(face)
{
	st->pc.face(ARG(0));
}

EXEC(label)
{
	// No operation
}

EXEC(goto)
{
	pair<int, int> p = st->pgm->labels[ARG(0)];
	if (p.first < 0)
		throw hippo_error("Skok na nedefinované návěští");
	st->pc.i = p.first;
	st->pc.j = p.second;
	st->pc.unstep();
}

EXEC(init)
{
	st->pc.dir_i = 1;
	st->pc.dir_j = 0;
}

EXEC(halt)
{
	throw hippo_halt();
}

EXEC(jump)
{
	st->pc.step();
}

EXEC(call)
{
	if (st->return_stack.size() >= MAX_STACK_DEPTH)
		throw hippo_error("Příliš mnoho vnořených CALLů");
	st->return_stack.push_back(st->pc);
	st->pc.turn(ARG(0));
}

EXEC(return)
{
	if (!st->return_stack.size())
		throw hippo_error("RETURN s prázdným zásobníkem");
	st->pc = st->return_stack.back();
	st->return_stack.pop_back();
}

EXEC(array)
{
	st->reg_new_array(ARG(0), RR(1));
}

EXEC(read)
{
	shared_ptr<array_state> a = RRA(1);
	int i = RR(2);
	if (i < 0 || i >= (int) a->x.size())
		throw hippo_error("Index pole mimo rozsah");
	WR(0, a->x[i]);
}

EXEC(write)
{
	shared_ptr<array_state> a = RRA(0);
	int i = RR(1);
	if (i < 0 || i >= (int) a->x.size())
		throw hippo_error("Index pole mimo rozsah");
	a->x[i] = RR(2);
}

// List of instructions

class insn_def insn_defs[] = {
#define I(s,d,n) { .syntax = s, .display = d, .exec = exec_##n },
#include "insn.h"
#undef I
};

#define NUM_INSNS (sizeof(insn_defs) / sizeof(insn_defs[0]))

// Parser

int insn_def::parse(const string s, insn *ins)
{
	int i = 0;
	int j = 0;
	int len = s.length();

	while (syntax[i]) {
		char c = syntax[i++];
		if (c >= 'a' && c <= 'z') {
			switch (c) {
				case 'r':
					if (j >= len)
						return 2;
					else if (s[j] >= 'a' && s[j] <= 'z')
						ins->add_arg(s[j++] - 'a');
					else
						return -1;
					break;
				case 'x':
					if (j >= len)
						return 2;
					else if (s[j] >= 'a' && s[j] <= 'z')
						ins->add_arg(s[j++] - 'a');
					else if (s[j] >= '0' && s[j] <= '9')
						ins->add_arg(s[j++] - '0' + 1000);
					else
						return -1;
					break;
				case 'l':
					if (j >= len)
						return 2;
					else if (s[j] >= 'a' && s[j] <= 'z')
						ins->add_arg(s[j++] - 'a');
					else
						return -1;
					break;
				case 'd':
					if (j >= len)
						return 2;
					else {
						switch (s[j++]) {
							case 'l': ins->add_arg(0); break;
							case 'r': ins->add_arg(1); break;
							case 'b': ins->add_arg(2); break;
							default: return -1;
						}
					}
					break;
				case 'a':
					if (j >= len)
						return 2;
					else {
						switch (s[j++]) {
							case 'n': ins->add_arg(0); break;
							case 'e': ins->add_arg(1); break;
							case 'w': ins->add_arg(2); break;
							case 's': ins->add_arg(3); break;
							default: return -1;
						}
					}
					break;
				default:
					return false;
			}
		} else {
			if (j >= len)
				return 1;
			else if (s[j] == tolower(c))
				j++;
			else
				return 0;
		}
	}

	return (j == len ? 3 : -1);
}

insn::insn()
{
	def = &insn_defs[0];	// NOP
	nargs = 0;
	for (int i=0; i < MAX_ARGS; i++)
		arg[i] = 0;
}

bool insn::parse(const string s)
{
	for (uint i=0; i < NUM_INSNS; i++)
		if (insn_defs[i].parse(s, this) == 3) {
			def = &insn_defs[i];
			return true;
		}
	return false;
}

int insn::partial_parse(const string s)
{
	bool had_partial = false;
	for (uint i=0; i < NUM_INSNS; i++) {
		int res = insn_defs[i].parse(s, this);
		if (res == 0)
			;
		else if (res == 1)
			had_partial = true;
		else {
			def = &insn_defs[i];
			return res;
		}
	}
	return had_partial ? 1 : 0;
}

string insn_def::show(const insn *ins)
{
	int ai = 0;
	string s = "";

	for (int i=0; syntax[i]; i++) {
		char c = syntax[i];
		switch (c) {
			case 'r':
			case 'l':
				if (ai < ins->nargs)
					s.push_back('a' + ins->arg[ai++]);
				else
					s.push_back('?');
				break;
			case 'x':
				if (ai < ins->nargs) {
					if (ins->arg[ai] >= 1000)
						s.push_back('0' + ins->arg[ai++] - 1000);
					else
						s.push_back('a' + ins->arg[ai++]);
				} else
					s.push_back('?');
				break;
			case 'd':
				if (ai >= ins->nargs) {
					s.push_back('?');
					break;
				}
				switch (ins->arg[ai++]) {
					case 0: s.push_back('l'); break;
					case 1: s.push_back('r'); break;
					case 2: s.push_back('b'); break;
					default: s.push_back('?');
				}
				break;
			case 'a':
				if (ai >= ins->nargs) {
					s.push_back('?');
					break;
				}
				switch (ins->arg[ai++]) {
					case 0: s.push_back('n'); break;
					case 1: s.push_back('e'); break;
					case 2: s.push_back('w'); break;
					case 3: s.push_back('s'); break;
					default: s.push_back('?');
				}
				break;
			default:
				s.push_back(tolower(c));
		}
	}

	return s;
}

string insn_def::show_pretty(const insn *ins)
{
	int ai = 0;
	string s = "";

	for (int i=0; display[i]; i++) {
		char c = display[i];
		switch (c) {
			case 'r':
			case 'l':
				if (ai < ins->nargs)
					s.push_back('a' + ins->arg[ai++]);
				else
					s.push_back('?');
				break;
			case 'x':
				if (ai < ins->nargs) {
					if (ins->arg[ai] >= 1000)
						s.push_back('0' + ins->arg[ai++] - 1000);
					else
						s.push_back('a' + ins->arg[ai++]);
				} else
					s.push_back('?');
				break;
			case 'd':
				if (ai >= ins->nargs) {
					s.push_back('?');
					break;
				}
				switch (ins->arg[ai++]) {
					case 0: s.append("left"); break;
					case 1: s.append("right"); break;
					case 2: s.append("back"); break;
					default: s.push_back('?');
				}
				break;
			case 'a':
				if (ai >= ins->nargs) {
					s.push_back('?');
					break;
				}
				switch (ins->arg[ai++]) {
					case 0: s.append("north"); break;
					case 1: s.append("east"); break;
					case 2: s.append("west"); break;
					case 3: s.append("south"); break;
					default: s.push_back('?');
				}
				break;
			default:
				s.push_back(tolower(c));
		}
	}

	return s;
}
