/*
 *	Hippogryf -- Program state
 *
 *	(c) 2018 Martin Mares <mj@ucw.cz>
 */

#include "hippogryf.h"

#include <fstream>

// Loading and saving

program::program()
{
}

void program::load(istream *in)
{
	int i = 0;
	while (in->peek() >= 0) {
		if (i >= NUM_ROWS)
			throw hippo_error(i, 0, "Příliš mnoho řádků");
		int j = 0;
		for (;;) {
			string s = "";
			int c = 0;
			for (;;) {
				c = in->get();
				if (c < 0)
					throw hippo_error(i, j, "Neočekávaný konec souboru");
				if (c == '\n' || c == '\t')
					break;
				s.push_back(c);
			}
			if (j >= NUM_COLS)
				throw hippo_error(i, j, "Příliš mnoho sloupců");
			if (s != "") {
				if (!ins[i][j].parse(s))
					throw hippo_error(i, j, "Neznámá instrukce");
			}
			if (c == '\n')
				break;
			j++;
		}
		i++;
	}
}

void program::load(const string filename)
{
	fstream fi(filename, fi.in);
	if (!fi.is_open())
		throw hippo_error("Nemohu načíst soubor");
	load(&fi);
}

void program::save(ostream *out)
{
	for (int i=0; i<NUM_ROWS; i++) {
		for (int j=0; j<NUM_COLS; j++) {
			if (j)
				*out << '\t';
			*out << ins[i][j].def->show(&ins[i][j]);
		}
		*out << endl;
	}
}

void program::save(const string filename)
{
	fstream fo(filename, fo.out | fo.trunc);
	if (!fo.is_open())
		throw hippo_error("Nemohu vytvořit soubor");
	save(&fo);
}

// Resolving of labels

void program::resolve_labels()
{
	for (int i=0; i<=26; i++)
		labels[i] = pair<int, int>(-1, -1);
	for (int i=0; i<NUM_ROWS; i++) {
		for (int j=0; j<NUM_COLS; j++) {
			insn *ii = &ins[i][j];
			if (ii->def->syntax[0] == 'L') {
				if (labels[ii->arg[0]].first >= 0)
					throw hippo_error(i, j, "Návěští definované vícekrát");
				labels[ii->arg[0]] = pair<int, int>(i, j);
			} else if (ii->def->syntax[0] == 'I') {
				if (labels[26].first >= 0)
					throw hippo_error(i, j, "INIT definovaný vícekrát");
				labels[26] = pair<int, int>(i, j);
			}
		}
	}
}

// Program state

void pc_state::step()
{
	i = (i + dir_i + NUM_ROWS) % NUM_ROWS;
	j = (j + dir_j + NUM_COLS) % NUM_COLS;
}

void pc_state::unstep()
{
	i = (i - dir_i + NUM_ROWS) % NUM_ROWS;
	j = (j - dir_j + NUM_COLS) % NUM_COLS;
}

void pc_state::face(int dir)
{
	switch (dir) {
		case 0: dir_i = -1; dir_j =  0; break;
		case 1: dir_i =  0; dir_j =  1; break;
		case 2: dir_i =  0; dir_j = -1; break;
		case 3: dir_i =  1; dir_j =  0; break;
		default: assert(0);
	}
}

void pc_state::turn(int dir)
{
	int di = dir_i, dj = dir_j;
	switch (dir) {
		case 0: dir_i = -dj; dir_j =  di; break;
		case 1: dir_i =  dj; dir_j = -di; break;
		case 2: dir_i = -di; dir_j = -dj; break;
		default: assert(0);
	}
}

bool state::reg_is_num(int i)
{
	return !reg_is_array(i);
}

bool state::reg_is_array(int i)
{
	assert(i >= 0 && i < 26);
	return array_reg[i] != NULL;
}

void state::reg_set_num(int i, int x)
{
	assert(i >= 0 && i < 26);
	if (array_reg[i])
		array_reg[i] = shared_ptr<array_state>();
	num_reg[i] = x;
}

void state::reg_set_array(int i, shared_ptr<array_state> a)
{
	assert(i >= 0 && i < 26);
	array_reg[i] = a;
}

void state::reg_new_array(int i, int len)
{
	assert(i >= 0 && i < 26);
	if (len < 0)
		throw hippo_error("Pokus o vytvoření pole záporné velikosti");
	if (len > MAX_ARRAY_LEN)
		throw hippo_error("Pokus o vytvoření příliš velkého pole");
	shared_ptr<array_state> a = make_shared<array_state>();
	a->x.resize(len);
	for (int j=0; j<len; j++)
		a->x[j] = 0;
	array_reg[i] = a;
}

int state::reg_get_num(int i)
{
	assert(i >= 0 && i < 26);
	if (!reg_is_num(i))
		throw hippo_error("Registr neobsahuje číslo");
	return num_reg[i];
}

shared_ptr<array_state> state::reg_get_array(int i)
{
	assert(i >= 0 && i < 26);
	if (!reg_is_array(i))
		throw hippo_error("Registr neobsahuje pole");
	return array_reg[i];
}

void state::show_regs()
{
	for (int i=0; i<26; i++) {
		cout << char('A' + i) << "=";
		if (reg_is_num(i))
			cout << reg_get_num(i);
		else {
			cout << "[";
			shared_ptr<array_state> a = reg_get_array(i);
			for (int j=0; j < (int) a->x.size(); j++) {
				if (j)
					cout << ", ";
				cout << a->x[j];
			}
			cout << "]";
		}
		cout << endl;
	}
}

// Execution

state::state(const program *_pgm)
{
	pgm = _pgm;

	if (pgm->labels[26].first < 0)
		throw hippo_error("INIT nenalezen");
	pc.i = pgm->labels[26].first;
	pc.j = pgm->labels[26].second;
	run_time = 0;
	debug = 0;

	for (int i=0; i<26; i++) {
		num_reg[i] = 0;
		array_reg[i] = NULL;
	}
}

void state::run_step()
{
	if (run_time >= MAX_RUN_TIME)
		throw hippo_error("Časový limit vypršel");
	const insn *ii = &pgm->ins[pc.i][pc.j];
	if (debug)
		cout << to_string(run_time) + " [" + to_string(pc.i) + "," + to_string(pc.j) + "] " + ii->def->show_pretty(ii) << endl;
	ii->def->exec(ii, this);
	run_time++;
	pc.step();
}

void state::run()
{
	for (;;) {
		try {
			run_step();
		} catch (hippo_error &e) {
			if (e.i < 0)
				throw hippo_error(pc.i, pc.j, e.what());
			else
				throw e;
		} catch (hippo_halt &e) {
			return;
		}
	}
}

void state::read_input(string in)
{
	ifstream f(in);
	if (!f.is_open())
		throw hippo_error("Nemohu otevřít vstup " + in);

	while (f.peek() >= 0) {
		int reg = f.get();
		if (!(reg >= 'A' && reg <= 'Z'))
			throw hippo_error("Chybný vstup: očekávám jméno registru");
		reg -= 'A';
		int ch = f.get();
		if (ch == '=') {
			if (f.peek() == '[') {
				f.get();
				shared_ptr<array_state> a = make_shared<array_state>();
				for (;;) {
					if (f.peek() == ']') {
						f.get();
						break;
					}
					if (f.peek() == ',') {
						f.get();
						continue;
					}
					int i;
					f >> i;
					a->x.push_back(i);
				}
				reg_set_array(reg, a);
			} else {
				int i;
				f >> i;
				reg_set_num(reg, i);
			}
		} else if (ch == '?') {
			if (f.peek() == '[') {
				f.get();
				if (f.get() != ']')
					throw hippo_error("Chybný vstup: pochybný dotaz");
				output_spec.push_back(pair<int,int>(reg, 1));
			} else
				output_spec.push_back(pair<int,int>(reg, 0));
		} else
			throw hippo_error("Chybný vstup: Očekávám = nebo ?");
		if (f.get() != '\n')
			throw hippo_error("Chybný vstup: Neukončený řádek");
	}
}

void state::write_output(string out)
{
	ofstream f(out, f.out | f.trunc);
	if (!f.is_open())
		throw hippo_error("Nemohu vytvořit výstup " + out);

	for (auto a: output_spec) {
		int reg = a.first;
		if (a.second) {
			if (!reg_is_array(reg))
				throw hippo_error(string("Špatný výstup!\nV registru ") + char('A'+reg) + " není pole");
			shared_ptr<array_state> as = reg_get_array(reg);
			bool first = true;
			for (int x: as->x) {
				if (!first)
					f << ' ';
				first = false;
				f << x;
			}
			f << endl;
		} else {
			if (!reg_is_num(reg))
				throw hippo_error(string("Špatný výstup!\nV registru ") + char('A'+reg) + " není číslo");
			f << reg_get_num(reg) << endl;
		}
	}
}
